/*
 *  gosm.h
 *  GOSM
 *
 *  Created by Ivica Aracic on 01.06.10.
 *  Copyright 2010 Ivica Aracic. All rights reserved.
 *
 */


#ifndef _GOSM_H_
#define _GOSM_H_

#include <vector>
#include <list>
#include <string>

using namespace std;

#define GOSM_DO_PROTOCOLL

#ifdef GOSM_DO_LOG
	#define GOSM_LOG(...) printf(__VA_ARGS__)
#else
	#define GOSM_LOG(...)
#endif

#ifdef GOSM_DO_PROTOCOLL
	#define GOSM_PROTOCOLL_FRAME(p) (p += "\n\n#; ")
	#define GOSM_PROTOCOLL_TRANSITION(p, t) (p += "T:"+(t).source->getStrId()+"->"+(t).target->getStrId()+"(from="+(t).inheritedFrom->getStrId()+"); ")
	#define	GOSM_PROTOCOLL(p, node, action) (p += node->getStrId() + action + string("; "))
	#define GOSM_PROTOCOLL_TF(p) (p += "<tf>.exec; ")
#else
	#define GOSM_PROTOCOLL_FRAME(p)
	#define GOSM_PROTOCOLL_TRANSITION(p, t)
	#define GOSM_PROTOCOLL(p, node, action)
	#define GOSM_PROTOCOLL_TF(p)
#endif

#define GOSM_ASSERT(exp) if(!(exp)) {GOSM_LOG("'%s' failed in %s:%d\n", #exp, __FILE__, __LINE__); exit(-1);}

namespace gosm {
	template <class T> class State;
	template <class T> class GuardCondition;
	template <class T> class TransitionFn;

	enum STATE_ID {
		STATE_ID_Inactive = -1,		// always. we separate init and inactive in order to enable reactivation (e.g. cloud/ufo in ccb)

		STATE_ID_Init = -2,		// always. we separate init and inactive in order to enable reactivation (e.g. cloud/ufo in ccb)
		// init/dead replace gameobjectlifecycle event
		STATE_ID_Dead = -3,		// always

		STATE_ID_ROOT = -4,

		STATE_ID_HISTORY = -5,

		STATE_ID_DEEP_HISTORY = -6
	};

	class StateMachineObject {
	public:
		virtual ~StateMachineObject() {}
	};



	/////////////////////////////////////////////////
	template <class T>
	class TransitionFn : public StateMachineObject {
	public:
		virtual void exec(T* go) {}
	};

	template <class T>
	class TransitionFnMemberFnPtr : public TransitionFn<T> {
	public:
		typedef void (T::*TransitionExecFn)();

	public:
		TransitionFnMemberFnPtr(TransitionExecFn execFn)
		: execFn(execFn) {}

		virtual void exec(T* go) {(go->*execFn)();}

	protected:
		TransitionExecFn execFn;
	};



	/////////////////////////////////////////////////
	template <class T>
	class Transition : public StateMachineObject {
	public:
		typedef State<T> State;
		typedef GuardCondition<T> GuardCondition;
		typedef TransitionFn<T> TransitionFn;

		Transition()
		: source(NULL), target(NULL), gc(NULL), transitionFn(NULL), inheritedFrom(NULL) {
		}

		Transition(State* source, State* target, GuardCondition* gc, TransitionFn* transitionFn, State* inheritedFrom)
		: source(source), target(target), gc(gc), transitionFn(transitionFn), inheritedFrom(inheritedFrom) {
		}

		bool isInherited() {return source != inheritedFrom;}

	public:
		State* source;
		State* target;
		GuardCondition* gc;
		TransitionFn* transitionFn;
		State* inheritedFrom;
	};



	/////////////////////////////////////////////////
	template <class T>
	class GuardCondition : public StateMachineObject {
	public:
		typedef State<T> State;
		virtual bool eval(T* go, State* ctx) = 0;
	};

	template <class T>
	class GuardConditionFnPtr : public GuardCondition<T> {
	public:
		typedef bool (*GuardConditionFn)(State<T>* ctx);

	public:
		GuardConditionFnPtr(GuardConditionFn fn) : fn(fn) {}

		virtual bool eval(T* go, State<T>* ctx) {return fn(ctx);}

	protected:
		GuardConditionFn fn;
	};

	template <class T>
	class GuardConditionMemberFnPtr : public GuardCondition<T> {
	public:
		typedef bool (T::*GuardConditionFn)(State<T>* ctx);

	public:
		GuardConditionMemberFnPtr(GuardConditionFn fn) : fn(fn) {}

		virtual bool eval(T* go, State<T>* ctx) {return (go->*fn)(ctx);}

	protected:
		GuardConditionFn fn;
	};

	template <class T>
	class GuardConditionBinaryBoolOp : public GuardCondition<T> {
	public:
		typedef bool (*CompFn)(bool arg1, bool arg2);

	public:
		GuardConditionBinaryBoolOp(CompFn fn, GuardCondition<T>* gc1, GuardCondition<T>* gc2)
		: fn(fn), gc1(gc1), gc2(gc2) {}

		virtual bool eval(T* go, State<T>* ctx) {return fn(gc1->eval(go, ctx), gc2->eval(go, ctx));}

	protected:
		CompFn fn;
		GuardCondition<T>* gc1;
		GuardCondition<T>* gc2;

	public:
		static bool orFn(bool arg1, bool arg2) {return arg1 || arg2;}
		static bool andFn(bool arg1, bool arg2) {return arg1 && arg2;}
	};

	template <class T>
	class GuardConditionUnaryBoolOp : public GuardCondition<T> {
	public:
		typedef bool (*CompFn)(bool arg);

	public:
		GuardConditionUnaryBoolOp(CompFn fn, GuardCondition<T>* gc)
		: fn(fn), gc(gc) {}

		virtual bool eval(T* go, State<T>* ctx) {return fn(gc->eval(go, ctx));}

	protected:
		CompFn fn;
		GuardCondition<T>* gc;

	public:
		static bool notFn(bool arg) {return !arg;}
	};



	/////////////////////////////////////////////////
	template <class T>
	class State : public StateMachineObject {
	public:
		typedef State<T> THIS;
		typedef GuardCondition<T> GuardCondition;
		typedef TransitionFn<T> TransitionFn;
		typedef Transition<T> Transition;

		typedef list<THIS*> ChildrenSeq;
		typedef vector<Transition> RuleSequence;

	public:
		State() {
			_initState(STATE_ID_ROOT, "ROOT", NULL, NULL);
		}

		State(int id, string const& strId, THIS* parent, float* timer = NULL) {
			_initState(id, strId, parent, timer);
		}

		void _initState(int id, string const& strId, THIS* parent, float* timer) {
			GOSM_ASSERT(parent != NULL || id == STATE_ID_ROOT);
			GOSM_LOG("STATE: %s [%s]\n", strId.c_str(), parent ? parent->getStrId().c_str() : "NULL");
			this->id = id;
			this->strId = strId;
			this->parent = parent;
			this->timer = timer;

			this->parent = parent;

			_H = _HH = NULL;

			if(parent) {
				parent->getChildren().push_back(this);
			}
		}

		virtual ~State() {
			GOSM_LOG("~State: %s\n", getStrId().c_str());

			children.clear();

			if(_H) delete _H;
			if(_HH) delete _HH;
		}

	protected:
		class PartialTransitionGcTf {
		public:
			PartialTransitionGcTf(THIS* s, GuardCondition* gc, TransitionFn* tf)
			: s(s), gc(gc), tf(tf) {
			}

			void operator>>(State& target) {
				s->addTransition(&target, gc, tf);
			}

			void operator<<(State& target) {
				target.addTransition(s, gc, tf);
			}

			State* s;
			GuardCondition* gc;
			TransitionFn* tf;
		};

		class PartialTransitionGc {
		public:
			PartialTransitionGc(THIS* s, GuardCondition* gc) : s(s), gc(gc) {}

			PartialTransitionGcTf operator/(TransitionFn& tf) {
				return PartialTransitionGcTf(s, gc, &tf);
			}

			void operator>>(THIS& target) {
				s->addTransition(&target, gc, NULL);
			}

			void operator<<(THIS& target) {
				target.addTransition(s, gc, NULL);
			}

			THIS* s;
			GuardCondition* gc;
		};

	public:
		PartialTransitionGc operator[] (GuardCondition& gc) {
			return PartialTransitionGc(this, &gc);
		}

		PartialTransitionGcTf operator/ (TransitionFn& tf) {
			return PartialTransitionGcTf(this, NULL, &tf);
		}

		void operator>> (THIS& target) {
			this->addTransition(&target, NULL, NULL);
		}

		void operator<< (THIS& target) {
			target.addTransition(this, NULL, NULL);
		}

		void addTransition(THIS* to,
						   GuardCondition* gc, TransitionFn* transFn
						   ) {
			GOSM_ASSERT(!isHistoryState());

			addTransition(to, gc, transFn, this);
		}

		void addTransition(
						   THIS* to,
						   GuardCondition* gc, TransitionFn* transFn,
						   State* inheritedFrom
						   ) {
			if(isHistoryState()) return;

			GOSM_LOG("RULE: %s --> %s\n", this->getStrId().c_str(), to->getStrId().c_str());
			rules.push_back(Transition(this, to, gc, transFn, inheritedFrom));

			// all children inherit parents transitions
			for(typename ChildrenSeq::iterator it=children.begin(); it!=children.end(); it++) {
				State* child = *it;
				child->addTransition(to, gc, transFn, inheritedFrom);
			}
		}


		// -----

		virtual void doExec(T* go, float dTime) {}
		virtual void doEnter(T* go) {}
		virtual void doExit(T* go) {}

		bool isHistoryState() {return id==STATE_ID_HISTORY || id==STATE_ID_DEEP_HISTORY;}
		virtual int getId() {return id;}
		virtual string& getStrId() {return strId;}
		virtual float getStateTimer() {return timer ? *timer : 0;}

		virtual State* getParent() {return parent;}
		virtual ChildrenSeq& getChildren() {return children;}
		virtual bool isRoot() {return parent==NULL;}
		virtual bool isLeaf() {return children.size() == 0;}

		Transition* findTransition(T* go) {
			// (-) go through all rules where current state is the source
			//     and check for guard conditions
			for(
				typename RuleSequence::iterator it = getRules().begin();
				it != getRules().end();
				it++
			) {
				Transition& t = *it;

				if(t.source == this) {
					if(t.gc == NULL || t.gc->eval(go, t.inheritedFrom)) {
						return &t;
					}
				}
			}

			return NULL;
		}

		RuleSequence& getRules() {return rules;}

		THIS& H() {
			if(!_H) {
				_H = new THIS(STATE_ID_HISTORY, strId+".H", this);
			}
			return *_H;
		}

		THIS& HH() {
			if(!_HH) {
				_HH = new THIS(STATE_ID_DEEP_HISTORY, strId+".H*", this);
			}
			return *_HH;
		}

		THIS* getInitState() {
			return children.size() > 0 ? children.front() : NULL;
		}

		int getNodeIdx() {return nodeIdx;}
		int getNodeCount() {return nodeCount;}
		int getNodeDepth() {return nodeDepth;}

		int calcNodeData(int nextIdx = 0, int depth = 0) {
			nodeIdx = nextIdx++;
			nodeDepth = depth;

			for(typename ChildrenSeq::iterator it=children.begin(); it!=children.end(); it++) {
				nextIdx = (*it)->calcNodeData(nextIdx, depth + 1);
			}

			nodeCount = nextIdx - nodeIdx;

			GOSM_LOG("[%10s] count=%d idx=%d depth=%d\n", getStrId().c_str(), nodeCount, nodeIdx, nodeDepth);

			return nextIdx;
		}

	protected:		// transition definition
		RuleSequence rules;

	protected:		// structure
		THIS* _H;
		THIS* _HH;

		THIS* parent;
		ChildrenSeq children;
		int id;
		string strId;
		float* timer;

		int nodeIdx;	// unique index in this hierarchy
		int nodeCount;	// 1 + all children
		int nodeDepth;
	};




	template <class T>
	class StateMemberFnPtr : public State<T> {
	public:
		typedef void (T::*StateEnterExitFn)();
		typedef void (T::*StateUpdateFn)(float dTime);

	public:
		StateMemberFnPtr(int id, const char* strId, State<T>* parent, float* timer=NULL, StateUpdateFn execFn=NULL, StateEnterExitFn enterFn=NULL, StateEnterExitFn exitFn=NULL)
		: State<T>(id, strId, parent, timer), execFn(execFn), enterFn(enterFn), exitFn(exitFn) {}

		virtual void doExec(T* go, float dTime) {if(execFn) (go->*execFn)(dTime);}
		virtual void doEnter(T* go) {if(enterFn) (go->*enterFn)();}
		virtual void doExit(T* go) {if(exitFn) (go->*exitFn)();}

	protected:
		StateUpdateFn execFn;
		StateEnterExitFn enterFn;
		StateEnterExitFn exitFn;
	};



	//////////////////////////////

	template <class T>
	class StateMachineDef {
	public:
		StateMachineDef() {}
		virtual ~StateMachineDef() {
			GOSM_LOG("~StateMachineDef\n");
			GOSM_LOG("\tdeleting %d smRefs\n", (int)smRefs.size());
			for (vector<StateMachineObject*>::iterator it=smRefs.begin(); it!=smRefs.end(); it++) {
				delete *it;
			}
		}

		State<T>* getRoot() {return &rootState;}

		void update(T* go, float dTime) {
			rootState.update(go, dTime);
		}

		template <class S>
		S* addSmRef(S* ref) {
			smRefs.push_back(ref); return ref;
		}

	protected:
		State<T> rootState;
		vector<StateMachineObject*> smRefs;

	public:
		static StateMachineDef* begin() {
			GOSM_ASSERT(current == NULL);
			current = new StateMachineDef();
			return current;
		}

		static void end() {
			GOSM_ASSERT(current != NULL);
			current->getRoot()->calcNodeData();
			current = NULL;
		}

		static StateMachineDef* getCurrent() {
			return current;
		}

	protected:
		static StateMachineDef* current;
	};

	template <class T>
	class StateMachine {
	protected:
		class RuntimeState {
			friend class StateMachine;
		public:
			RuntimeState() {
				history = NULL;
				currentTimer = 0;
			}

			void update(float dTime) {
				currentTimer -= dTime;
				if(currentTimer < 0) currentTimer = 0;
			}

			float getTimer() {return currentTimer;}
			void setTimer(float timer) {currentTimer = timer;}
			bool isTimerTimeout() {return currentTimer <= 0;}

			void setHistory(RuntimeState* history) {this->history = history;}
			RuntimeState* getHistory() {return history;}

			RuntimeState* getDeepHistory() {
				if(history) {
					if(history->getHistory() == NULL) {
						return history;
					}
					else {
						return history->getDeepHistory();
					}
				}
				else {
					return NULL;
				}
			}

		protected:
			State<T>* s;
			RuntimeState* history;
			float currentTimer;
		};
	public:
		StateMachine(StateMachineDef<T>* def)
		: def(def) {
			execStackActualCounter = execStackDesiredCounter = 0;
			createRuntime();
		}

		virtual ~StateMachine() {
			GOSM_LOG("~StateMachine\n");
			delete[] runtimeStates;
		}

		void update(T* go, float dTime) {
			GOSM_PROTOCOLL_FRAME(PROTOCOLL);

			// empty stack
			if(execStackActualCounter == 0) {
				execStackBuildDesired(def->getRoot());
				execStackProcessExit(go);
				execStackProcessEnter(go);
			}

			Transition<T>* t = NULL;

			int stackIdx = 0;
			while (stackIdx < execStackActualCounter && !t) {
				t = execStack[stackIdx++].actualState->findTransition(go);
			}

			execStackUpdate(go, dTime);

			if(t) {
				GOSM_PROTOCOLL_TRANSITION(PROTOCOLL, *t);

				execStackBuildDesired(t->target);
				//execStackPrint();

				execStackProcessExit(go);
				//execStackPrint();

				GOSM_PROTOCOLL_TF(PROTOCOLL);
				if(t->transitionFn) {
					t->transitionFn->exec(go);
				}

				execStackProcessEnter(go);
				//execStackPrint();
			}
		}

		RuntimeState* getRuntimeState(State<T>* state) {
			int nodeIdx = state->getNodeIdx();

			if(state && nodeIdx>=0 && nodeIdx<runtimeStatesCount) {
				runtimeStates[nodeIdx].s = state;
				return &runtimeStates[nodeIdx];
			}

			return NULL;
		}

	protected:
		void createRuntime() {
			runtimeStatesCount = def->getRoot()->getNodeCount();

			runtimeStates = new RuntimeState[runtimeStatesCount];
		}

	protected:
		StateMachineDef<T>* def;
		int runtimeStatesCount;
		RuntimeState *runtimeStates;

#ifdef GOSM_DO_PROTOCOLL
	public:
		string PROTOCOLL;
#endif

	public:
		State<T>* getExecStackTop() {
			if(execStackActualCounter > 0) {
				return execStack[execStackActualCounter-1].actualState;
			}

			return NULL;
		}

	protected:
		enum CONSTS {MAX_STACK_DEPTH = 8};

		class ExecStackEntry {
		public:
			ExecStackEntry() : actualState(NULL), desiredState(NULL) {}

			bool checkBothEqual() {return actualState == desiredState;}

			State<T>* actualState;
			State<T>* desiredState;
		};

		ExecStackEntry execStack[MAX_STACK_DEPTH];

		int execStackActualCounter;
		int execStackDesiredCounter;

		void execStackProcessExit(T* go) {
			int idx = execStackActualCounter-1;
			while (
				idx >= 0
				&& (idx >= execStackDesiredCounter
					|| !execStack[idx].checkBothEqual())
			) {
				State<T>* s = execStack[idx].actualState;

				GOSM_PROTOCOLL(PROTOCOLL, s, ".exit");
				s->doExit(go);

				// remember history
				if(s->getParent()) {
					getRuntimeState(s->getParent())->setHistory(getRuntimeState(s));
				}

				idx--;
			}

			execStackActualCounter = idx + 1;
		}

		void execStackProcessEnter(T* go) {
			while (execStackActualCounter < execStackDesiredCounter) {
				int idx = execStackActualCounter++;

				execStack[idx].actualState = execStack[idx].desiredState;
				GOSM_PROTOCOLL(PROTOCOLL, execStack[idx].actualState, ".enter");

				getRuntimeState(execStack[idx].actualState)->setTimer(execStack[idx].actualState->getStateTimer());

				execStack[idx].actualState->doEnter(go);
			}

			execStackDesiredCounter = 0;
		}

		void execStackUpdate(T* go, float dTime) {
			for(int i=0; i<execStackActualCounter; i++) {
				GOSM_PROTOCOLL(PROTOCOLL, execStack[i].actualState, ".exec");

				getRuntimeState(execStack[i].actualState)->update(dTime);
				execStack[i].actualState->doExec(go, dTime);
			}
		}

		void execStackBuildDesired(State<T>* n) {
			// check if history
			if(n->isHistoryState()) {
				RuntimeState* h = NULL;

				if(n->getId() == STATE_ID_HISTORY) {
					h = getRuntimeState(n->getParent())->getHistory();
				}
				else if(n->getId() == STATE_ID_DEEP_HISTORY) {
					h = getRuntimeState(n->getParent())->getDeepHistory();
				}

				if(h) {
					n = h->s;
				}
				else {
					n = n->getParent()->getInitState();
				}
			}

			execStackDesiredCounter = n->getNodeDepth()+1;
			GOSM_ASSERT(execStackDesiredCounter <= MAX_STACK_DEPTH);

			State<T>* _n = n;
			do {
				int idx = _n->getNodeDepth();

				GOSM_ASSERT(idx >= 0 && idx < execStackDesiredCounter);

				execStack[idx].desiredState = _n;

				_n = _n->getParent();
			} while(_n != NULL);

			// from here on do init with default
			State<T>* initChild = n->getInitState();
			while (execStackDesiredCounter < MAX_STACK_DEPTH && initChild) {
				execStack[execStackDesiredCounter++].desiredState = initChild;

				initChild = initChild->getInitState();
			}
		}

#ifdef GOSM_DO_LOG
		void execStackPrint() {
			GOSM_LOG("    %d %d\n", execStackActualCounter, execStackDesiredCounter);
			for(int i=0; i<MAX_STACK_DEPTH; i++) {
				GOSM_LOG("%d. %10s %10s\n", i,
						 i<execStackActualCounter?execStack[i].actualState->getStrId().c_str():"NULL",
						 i<execStackDesiredCounter?execStack[i].desiredState->getStrId().c_str():"NULL");
			}
		}
#endif

	public:
		bool gcTimeout(State<T>* ctx) {return getRuntimeState(ctx)->isTimerTimeout();}
	};


	template <class T> GuardCondition<T>& operator&& (GuardCondition<T>& gc1, GuardCondition<T>& gc2) {
		return *(StateMachineDef<T>::getCurrent()->addSmRef(new GuardConditionBinaryBoolOp<T>(GuardConditionBinaryBoolOp<T>::andFn, &gc1, &gc2)));
	}
	template <class T> GuardCondition<T>& operator|| (GuardCondition<T>& gc1, GuardCondition<T>& gc2) {
		return *(StateMachineDef<T>::getCurrent()->addSmRef(new GuardConditionBinaryBoolOp<T>(GuardConditionBinaryBoolOp<T>::orFn, &gc1, &gc2)));
	}
	template <class T> GuardCondition<T>& operator! (GuardCondition<T>& gc1) {
		return *(StateMachineDef<T>::getCurrent()->addSmRef(new GuardConditionUnaryBoolOp<T>(GuardConditionUnaryBoolOp<T>::notFn, &gc1)));
	}


	template <class T> StateMachineDef<T>* StateMachineDef<T>::current = NULL;

	// ------------------------

#define gc(_gc) *(gosm::StateMachineDef<THIS>::getCurrent()->addSmRef(new gosm::GuardConditionMemberFnPtr<THIS>(&THIS::gc##_gc)))
#define tf(_tf) *(gosm::StateMachineDef<THIS>::getCurrent()->addSmRef(new gosm::TransitionFnMemberFnPtr<THIS>(&THIS::tf##_tf)))

#define GOSM_ID(name) STATE_ID_##name, #name

	// ------------------------

}

#endif
